
PS1 = ">>> "
PS2 = "... "
view_detail = ('id', 'title', 'meta_type', 'mod_time')

import cmd, sys, string
import getopt, getpass
import xmlrpclib
from Support import BasicAuthTransport

try:
    from client_help import help_text, zoql_help
except:
    help_text = ''
    zoql_help = {}

try:
    import readline
except:
    pass

class ZOQLClient(cmd.Cmd):

    def __init__(self, url, username, password):
	''' setup '''
        self.line = ''
	self.prompt = PS1
	self.server = None

        if url:
            self._connect(url, username, password)


    def do_shell_open(self, line):
	''' '''
        options = string.split(string.strip(line), ' ')
        url = options[0]

        username = ''
        if len(options) >= 2:
            username = options[1]

        password = ''
        if len(options) >= 3:
            password = options[2]

        self._connect(url, username, password)


    def do_shell_close(self, line):
        self._disconnect()


    def do_shell_exit(self, line):
        return 1


    def do_shell(self, line):
        line_start = string.lower(string.split(string.strip(line), ' ')[0])
        if line_start in ('open', 'connect'):
            return self.do_shell_open(line[len(line_start):])
        elif line_start in ('close', 'disctonnect'):
            return self.do_shell_close(line)
        elif line_start in ('exit', 'quit', 'q'):
            return self.do_shell_exit(line)

    def do_help(self, line):
        ''' '''
        if string.strip(line) == '':
            print string.join(zoql_help.keys(), '\n')
        else:
            statement = string.strip(line)
            statement = string.upper(statement)
            if zoql_help.has_key(statement):
                print zoql_help[statement]
            else:
                print "Statement was not found. Your choices:"
                print string.join(zoql_help.keys(), '\n')


    def default(self, line):
        self.line += line + ' \n'
        if string.strip(line)[-1] == ';':
            self.run_query(self.line)
            self.line = '' 
            self.prompt = PS1
        else:
            self.prompt = PS2
            

    def preloop(self):
	''' hello '''
        print
	print "ZOQLClient 0.1.0: A ZOQL Client for Python"
        print "(You can leave the shell by typing '!q'. For help use '?'.)"
        print

    def postloop(self):
	''' hello '''
        self._disconnect()


    def _connect(self, url, username, password):
        ''' '''
        if not username:
            username = raw_input('Username: ')

        if not password:
            password = getpass.getpass()
        
        if url[:7] != 'http://':
            url = 'http://' + url

        self.server = xmlrpclib.Server(url, transport = BasicAuthTransport(username, password))
        self.url = url[7:]

        try:
            self.server.index_html()
            print "Connected to %s" %self.url
        except:
            self.server = None
            self.url = ''
            print "Could not connect to server %s" %self.url

        if self.server is not None:
            try:
                self.server.manage_addProduct.ZOQLMethod.manage_addZOQLMethod('zoql_server',
                                                                              'ZOQL Server component for client')
            except xmlrpclib.ProtocolError:
                pass
            except xmlrpclib.Fault:
                print "Notice: The zoql_server ZOQL Method exists already on the server." 


    def _disconnect(self):
        ''' '''
        if self.server is not None:
            self.server.manage_delObjects(['zoql_server'])
            print "Disconnected from %s" % self.url
            self.server = None
            self.url = ''
        else:
            print "Already disconnected!"
        

    def run_query(self, command):
	''' runs a command, changes path '''

        if self.server is None:
            print "Not connected!"
            return

        try:
            result = self.server.zoql_server.execute_xmlrpc(command)
        except Exception, exp:
            result = {'status': 1, 'values': exp}

        self._pretty_output(result)
        

    def _get_max_length(self, result):
        ''' '''
        lengths = {}
        keys = result[0].keys()
        for key in keys:
            lengths[key] = len(key) + len(key)%2

        for object in result:
            for key in keys:
                if len(str(object[key])) > lengths[key]:
                    lengths[key] = len(str(object[key])) + len(str(object[key]))%2

        return lengths
    

    def _pretty_output(self, result):
        ''' '''
        if result['status']:
            print "*** An Error occured ***"
            try:
                print result['values']['msg']
            except:
                print result['values']

        else:
            if result['values']:
                lengths = self._get_max_length(result['values'])

                print 
                
                line1 = '|'
                line2 = '+'
                for col in view_detail:
                    space = (lengths[col]-len(col)+2)/2
                    line1 += ' '*space + col + ' '*(space+len(col)%2) + '|'
                    line2 += '-'*(2*space+len(col)+len(col)%2) + '+'

                print line2
                print line1
                print line2

                for object in result['values']:
                    line = '|'
                    for col in view_detail:
                        space = (lengths[col]-len(str(object[col]))+1)
                        line += ' '+str(object[col]) + ' '*space + '|'

                    print line

                print line2
                print
        

def parse_options():
    # parse commandline options

    url = ''
    username = ''
    password = ''
    infile = None
    
    try:
        optlist, args = getopt.getopt(sys.argv[1:], "h:u:p:f:", [
            'help', 'user=', 'host=', 'password=', 'infile='])
    except getopt.error, message:
        sys.stderr.write(message)
        sys.exit(1)
    
    for option, param in optlist:
        if option in ('--help',):
            print help_text
            sys.exit(0)
            
        if option in ('-u', '--user'): username = param
        if option in ('-h', '--host'): url = 'http://' + param
        if option in ('-f', '--infile'):
            try:
                infile = open(param, "r")
            except:
                print "Could not open %s" % param
                sys.exit(1)
                
        if option in ('-p', '--password'): password = param

    return url, username, password, infile


if __name__ == '__main__':

    url, username, password, infile = parse_options()
    
    shell = ZOQLClient(url, username, password)
        
    if infile:
        shell.run_query(infile.read())
        shell._disconnect()
    else:
        try:
            shell.cmdloop()
        except KeyboardInterrupt:
            sys.exit()
