# This code is copyrighted by Stephan Richter and
# Thomas Foerster in September, 2001.
#
# written by Stephan Richter
# Last updated: October 4, 2001
import string
import CatalogOperators


def get_identifier_name(elem):
    if type(elem).__name__ == 'instance':
        return elem.attr
    else:
        return elem


class Compare:

    def __init__(self, orderby):
        self.orderby = orderby

    def __call__(self, x, y):

        for prop in self.orderby:
            if prop['property'] != 'id':
                result = cmp(getattr(x, prop['property'], None), getattr(y, prop['property'], None))                
            else:
                result = cmp(x.getId(), y.getId())                

            if string.lower(prop['direction']) == 'desc':
                result = -result
            if result != 0: break

        return result


class Executer:


    def do_select(self, base_object, selected_meta_types=[], searched_containers=[],
                  conditions=None, orderby=[], limit={}, recursive=0, use_indexes=1):
        ''' '''

        result = []

        # First task of business: Check whether we got some Catalogs to work with!
        indexes = find_indexes(base_object)

        if indexes and use_indexes:
            # Convert the operators to catalog operators
            conditions = CatalogOperators.convert_operator(conditions)

            # add a condition to just look for certain meta_types
            if selected_meta_types and string.lower(selected_meta_types[0]) not in ('all', '*'):
                meta_type_cond = CatalogOperators.Equal('meta_type', selected_meta_types) 
                if conditions is not None:
                    conditions = CatalogOperators.And(conditions, meta_type_cond)
                else:
                    conditions = meta_type_cond

            # for every index that was found, do the search
            for index in indexes:
                # Interesting, sometimes we do not have any conditions!
                if conditions:
                    result_brains = conditions.execute(index)
                else:
                    result_brains = index()

                # Convert the resulting brains to actual objects; yeah!
                for brain in result_brains:
                    object = index.getobject(brain.data_record_id_)
                    result.append(object)

        else:
            # Okay, we have no index, so we simply walk the tree.

            # Check whether we have a legal base object, and let's make it the
            # seach base
            if base_object:
                search_base = base_object
                if hasattr(base_object, 'aq_base'):
                    search_base = base_object.aq_base
            else:
                return result

            # Well, we should be able to grep for the objects
            if not hasattr(search_base, 'objectValues'):
                return result

            if not searched_containers:
                # Simply look for all objects in the search base, since no specific
                # objects were specified
                try:
                    objects = base_object.objectValues()
                except:
                    return result
            else:
                # Well, look only in these objects that were selected
                objects = []
                for container_path in searched_containers:
            
                    container_path = get_identifier_name(container_path)
                    container = base_object.unrestrictedTraverse(string.replace(container_path, '.', '/'))
                    
                    objects = objects + list(container.objectValues())

            # This short cut saves some time later...
            try:
                add_result = result.append
            except:
                raise AttributeError, 'result'

            
            for object in objects:
            
                # Has anyone a clue, why this is done? And what does dflag stand for?
                dflag = 0
                if hasattr(object, '_p_changed') and (object._p_changed == None):
                    dflag = 1

                # Get the pure object, since it can come to complications otherwise
                if hasattr(object, 'aq_base'):
                    object_base = object.aq_base
                else:
                    object_base = object

                # The interesting stuff: Apply the conditions on the object to find out
                # whether it fits the query
                if ( ( not selected_meta_types or object.meta_type in selected_meta_types or
                       string.lower(selected_meta_types[0]) in ('all', '*')) and
                     ( not conditions or (conditions and conditions.execute(object)) ) ):
                    add_result(object)
                    dflag = 0

                # Let's look inside the object as well
                if recursive and hasattr(object, 'objectValues'):
                    result += self.do_select(object, selected_meta_types, [], conditions,
                                             [], {}, recursive)

                # Do you guys know what that means?
                if dflag:
                    object._p_deactivate()

        # Let's order the result, based on the information passed
        if orderby:
            result.sort(Compare(orderby))

        # Mmmh, it would be so much more efficient to do the limit before sorting,
        # but we can't do that of course...
        if limit:
            end = limit['start'] + limit['size']
            if end >= len(result):
                end = len(result)
            result = result[limit['start']:end]

        return result


    def do_create_product(self, base_object, prod_id, prod_type='ZCLASS'):
        ''' '''

        if prod_type == 'PYTHON':
            print 'Not yet implemented'

        elif prod_type == 'ZCLASS':
            base_object.Control_Panel.Products.manage_addProduct(prod_id, prod_id+' product')
            return 1

        return 0
    

    def do_create_property(self, sheet, property):
        '''Create a property from the given information; return None'''
        id = property['name']
        prop_type = property['type']
        default = property['default']
        
        sheet.manage_addProperty(id, default, prop_type)


    def do_create_propertysheet(self, class_instance, sheet):
        '''Create a propertysheet from the parsed information'''
        id = sheet['name']
        title = id + " Property Sheet" 
        
        class_instance.propertysheets.common.manage_addCommonSheet(id, title)

        property_sheet = class_instance.propertysheets.common[id]
        for property in sheet['props']:
            self.do_create_property(property_sheet, property)
            

    def do_create_zclass(self, base_object, prod_id, class_id, attributes=[], options=[],
                         propertysheets=[], bases=[]):
        """ Create a ZClass """
        
        product = base_object.Control_Panel.Products[prod_id]

        option_dict = {'create_factory': 0, 'zope_object': 0}
        for option in options:
            if get_identifier_name(option.lhs) in option_dict.keys():
                option_dict[get_identifier_name(option.lhs)] = option.rhs

        meta_type = class_id
        for attribute in attributes:
            if get_identifier_name(attribute.lhs) == 'meta_type':
                meta_type = attribute.rhs
        
        product.manage_addZClass(class_id, '', bases, meta_type,
                                 option_dict['create_factory'], None, option_dict['zope_object'])

        class_instance = product[class_id]

        for attr in attributes:
            attr.execute(class_instance)

        if propertysheets is not None:
            for sheet in propertysheets:
                self.do_create_propertysheet(class_instance, sheet)


    def do_create_index(self, base_object, name, selected_meta_types,
                        index_list=[], conditions=[], recursive=0):

        base_object.manage_addProduct['ZCatalog'].manage_addZCatalog(name, 'ZOQL Index')

        catalog = base_object[name]

        for ic in ('title', 'bobobase_modification_time'):
            catalog.delColumn(ic)
            catalog.delIndex(ic)

        catalog.delColumn('summary')
        catalog.delIndex('PrincipiaSearchSource')

        for new_index in index_list:
            if new_index['name'] not in catalog.indexes():
                index_type = string.capitalize(new_index['type']) + 'Index'
                catalog.addIndex(new_index['name'], index_type)

        objects = self.do_select(base_object, selected_meta_types, [], conditions, [], {}, recursive, 0)

        for object in objects:
            catalog.catalog_object(object)
        

    def do_alter_class_add(self, base_object, class_ref, object_type, object_ref):

        if string.lower(object_type) == 'method':
            raise NotImplemented

        name = string.replace(object_ref['name'], '/', '.')

        if len(string.split(name, '.')) == 2:
            class_object = base_object.unrestrictedTraverse('/Control_Panel/Products/'+
                                                            string.replace(class_ref, '.', '/'))

            sheet, object_ref['name'] = string.split(name, '.')
            self.do_create_property(class_object.propertysheets.common[sheet], object_ref)

        elif len(string.split(name, '.')) == 1:
            class_object = base_object.unrestrictedTraverse(string.replace(class_ref, '.', '/'))

            self.do_create_property(class_object, object_ref)

        

    def do_alter_class_rename(self, base_object, class_ref, old_name, new_name):

        old_name = string.replace(old_name, '/', '.')

        if len(string.split(old_name, '.')) == 2:
            class_object = base_object.unrestrictedTraverse('/Control_Panel/Products/'+
                                                            string.replace(class_ref, '.', '/'))
            sheet_id, old_id = string.split(old_name, '.')
            sheet = class_object.propertysheets.common[sheet_id]
            
        if len(string.split(old_name, '.')) == 1:
            class_object = base_object.unrestrictedTraverse(string.replace(class_ref, '.', '/'))
            old_id = old_name
            sheet = class_object

        prop_type = sheet.getPropertyType(old_id)
        prop_value = sheet.getProperty(old_id)

        # delete ...
        sheet._delProperty(old_id)

        # Add the property again ...
        new_name = string.replace(new_name, '/', '.')
        new_name = string.split(new_name, '.')[-1]

        sheet.manage_addProperty(new_name, prop_value, prop_type)
        

    def do_alter_class_delete(self, base_object, class_ref, property_ids):

        try:
            class_object = base_object.unrestrictedTraverse('/Control_Panel/Products/'+
                                                            string.replace(class_ref, '.', '/'))
            
            for id in property_ids:
                id = string.replace(id, '/', '.')
                sheet, prop = string.split(id, '.')
                class_object.propertysheets.common[sheet]._delProperty(prop)

        except:
            class_object = base_object.unrestrictedTraverse(string.replace(class_ref, '.', '/'))

            for id in property_ids:
                class_object._delProperty(id)


    def do_insert(self, base_object, meta_type, container_path, assignments):

        container = base_object.unrestrictedTraverse(string.replace(container_path, '.', '/'))
        
        # somewhere we need to find the id
        id = None
        for assignment in assignments:
            if get_identifier_name(assignment.lhs) == 'id':
                id = assignment.rhs
        
        if id is None:
            raise 'NoIdError', 'No Id was specified!'

        klass = None
        for entry in base_object.all_meta_types():
            if entry['name'] == meta_type:
                if entry.has_key('instance'):
                    klass = entry['instance']
                else:
                    for zclass in base_object.Control_Panel.Products[entry['product']].objectValues():
                        if zclass.meta_type == meta_type:
                            klass = zclass
                break

        instance = klass(id)
        instance._setId(id)
        container._setObject(id, instance)

        object = container[id]
        
        for assignment in assignments:
            if get_identifier_name(assignment.lhs) != 'id':
                assignment.execute(object)


    def do_update(self, base_object, selected_meta_types=[], searched_containers=[],
                  assignments=[], conditions=[], recursive=1):

        for container_path in searched_containers:

            container_path = get_identifier_name(container_path)
            container = base_object.unrestrictedTraverse(string.replace(container_path, '.', '/'))

            for object in container.objectValues(selected_meta_types):
                if not conditions or (conditions and conditions.execute(object)):
                    for assignment in assignments:
                        assignment.execute(object)

            if recursive:
                for object in container.objectValues():
                    if hasattr(object, 'objectValues'):
                        self.do_update(object, selected_meta_types, assignments,
                                       conditions, recursive)

    
    def do_delete(self, base_object, selected_meta_types=[],
                  searched_containers=[], conditions=None, recursive=1):

        for container_path in searched_containers:

            container_path = get_identifier_name(container_path)
            container = base_object.unrestrictedTraverse(string.replace(container_path, '.', '/'))

            for object in container.objectValues(selected_meta_types):
                if not conditions or (conditions and conditions.execute(object)):
                    container.manage_delObjects([object.getId()])
            
            if recursive:
                for object in container.objectValues():
                    if hasattr(object, 'objectValues'):
                        result += self.do_delete(self, object, selected_meta_types, searched_containers,
                                                 conditions, recursive)
            


def find_indexes(base_object):

    indexes = []
    for catalog in base_object.objectValues('ZCatalog'):
        if catalog.title == 'ZOQL Index':
            indexes.append(catalog)

    return indexes
